package ru.dbotthepony.mc.otm.core

import net.minecraft.core.BlockPos
import net.minecraft.core.Direction
import net.minecraft.core.Vec3i
import net.minecraft.world.phys.Vec3
import kotlin.math.sin
import kotlin.math.cos
import kotlin.math.acos

typealias Vector = Vec3

val VECTOR_UP = Vector(0.0, 1.0, 0.0)
val VECTOR_DOWN = Vector(0.0, -1.0, 0.0)
val VECTOR_BACK = Vector(0.0, 0.0, 1.0)

fun BlockPos.asVector(): Vector {
	return Vector(x + 0.5, y + 0.5, z + 0.5)
}

fun Vector.asAngle(): Angle {
	val norm = normalize()

	if (norm.x < 0) {
		return Angle(acos(norm.dot(VECTOR_UP)) - Angle.PI_HALF, -acos(Vector(norm.x, 0.0, norm.z).normalize().dot(VECTOR_BACK)))
	}

	return Angle(acos(norm.dot(VECTOR_UP)) - Angle.PI_HALF, acos(Vector(norm.x, 0.0, norm.z).normalize().dot(VECTOR_BACK)))
}

operator fun BlockPos.plus(direction: Vec3i): BlockPos = this.offset(direction)
operator fun BlockPos.plus(direction: Direction): BlockPos = this.offset(direction.normal)
operator fun BlockPos.minus(direction: Vec3i): BlockPos = this.subtract(direction)
operator fun BlockPos.minus(direction: Direction): BlockPos = this.subtract(direction.normal)

operator fun Vec3i.plus(direction: Vec3i): Vec3i = this.offset(direction)
operator fun Vec3i.plus(direction: Direction): Vec3i = this.offset(direction.normal)
operator fun Vec3i.minus(direction: Vec3i): Vec3i = this.subtract(direction)
operator fun Vec3i.minus(direction: Direction): Vec3i = this.subtract(direction.normal)

operator fun Vector.plus(direction: Vector): Vector = this.add(direction)
operator fun Vector.minus(direction: Vector): Vector = this.subtract(direction)

operator fun Vector.times(v: Double): Vector = this.multiply(v, v, v)

fun Vector.left() = asAngle().left() * this.length()
fun Vector.right() = asAngle().right() * this.length()
fun Vector.up() = asAngle().up() * this.length()
fun Vector.down() = asAngle().down() * this.length()

fun Vector.rotateAroundAxis(axis: Vector, rotation: Double): Vector {
	return this * cos(rotation) + axis.cross(this) * sin(rotation) + axis * axis.dot(this) * (1 - cos(rotation))
}

operator fun Vector.component1() = x
operator fun Vector.component2() = y
operator fun Vector.component3() = z

data class Angle(val pitch: Double = 0.0, val yaw: Double = 0.0, val roll: Double = 0.0) {
	fun forward(): Vector {
		val yaw = yaw - PI_HALF
		val x = cos(yaw) * cos(pitch)
		val y = -sin(pitch)
		val z = sin(yaw) * cos(pitch)

		return Vector(x, y, z)
	}

	fun right(): Vector {
		val pitch = 0.0
		val x = cos(yaw) * cos(pitch)
		val y = -sin(pitch)
		val z = sin(yaw) * cos(pitch)

		return Vector(x, y, z)
	}

	fun left(): Vector {
		val yaw = yaw + PI_HALF
		val pitch = 0.0
		val x = cos(yaw) * cos(pitch)
		val y = -sin(pitch)
		val z = sin(yaw) * cos(pitch)

		return Vector(x, y, z)
	}

	fun up(): Vector {
		val yaw = yaw - PI_HALF
		val pitch = pitch - PI_HALF
		val x = cos(yaw) * cos(pitch)
		val y = -sin(pitch)
		val z = sin(yaw) * cos(pitch)

		return Vector(x, y, z)
	}

	fun down(): Vector {
		val yaw = yaw - PI_HALF
		val pitch = pitch + PI_HALF
		val x = cos(yaw) * cos(pitch)
		val y = -sin(pitch)
		val z = sin(yaw) * cos(pitch)

		return Vector(x, y, z)
	}

	companion object {
		const val PI_HALF = Math.PI / 2.0

		fun degrees(pitch: Double, yaw: Double, roll: Double): Angle {
			return Angle(Math.toDegrees(pitch), Math.toDegrees(yaw), Math.toDegrees(roll))
		}
	}
}
